/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* $Id: quad-interface.h,v 1.32 2020/07/25 09:55:11 zlb Exp $
 *
 * Numerical quadrature on an element containing a smooth interface */

#ifndef PHG_QUAD_INTERFACE_H

#ifdef __cplusplus
extern "C" {
#endif

typedef struct QI_OPTIONS_ {
    FLOAT EPS;
    FLOAT threshold;	/* max a(w) allowed */

    /* whether use one of the face normals as the direction nu */
    BOOLEAN use_face_normal;

    /* The following flags control using shortcuts (faster but less accurate) */
    BOOLEAN shortcut_t;
    BOOLEAN shortcut_s;

    int adapt;

    int subdiv_type;
    INT subdiv_limit;	/* maximum recursion depth */
    INT subdiv_level0;	/* initial subdivision level */

    int newton;
    INT newton_maxits;	/* maximum Newton iterations */
    INT newton_porder;	/* polynomial order for initial solution */

    INT quad_n1;	/* nbr of intervals for get_quad_1D */
    INT quad_n2;	/* nbr of intervals/direction for get_quad_2D */

    /* debuging options */
    INT graph_t;
    INT graph_s;
    INT graph_w;

    FLOAT dbg_w;	/* <-1: auto, [-1,1]: specified, >1.0: random */
    INT   dbg_wN;	/* nits for 0.618 search, 0: middle point */

    FLOAT dbg_value_t; /* in [0,1] => debug with this t value */
    FLOAT dbg_value_s; /* in [0,1] => debug with this s value */

    BOOLEAN dbg_roots;

    BOOLEAN dbg_t;
    BOOLEAN dbg_s;
    BOOLEAN dbg_elem;
    BOOLEAN dbg_vtk;

    BOOLEAN dbg_rule_data;	/* debug computed rule data */

    BOOLEAN show_recursions;

    BOOLEAN show_vectors;
    BOOLEAN show_intervals_w;
    BOOLEAN show_intervals_s;
    BOOLEAN show_intervals_t;
} QI_OPTIONS;
extern QI_OPTIONS _phg_quad_interface_options;

/*-------------------------- Common functions --------------------------*/

typedef void (*QUAD_1D_FUNC)(FLOAT x, FLOAT *res, const FLOAT wgt, void *ctx);

void phgQuad1D(QUAD_1D_FUNC f, int dim, FLOAT a, FLOAT b, QUAD *quad,
		FLOAT *res, void *ctx);

#define RULE_HEADER	1
#define SetRuleHeader(rule, npoints, proj, type) \
	rule[0] = (FLOAT)npoints * 10.0 + (FLOAT)(proj + type * 5)

int phgQuadInterfaceApplyRule(void *func, int dim, const FLOAT *rule_data,
			      FLOAT *res);

int phgQuadInterfaceDumpRule_(const FLOAT *rule_data, FILE *fp,
			      DOF_PROJ *proj, int *type);
#define phgQuadInterfaceDumpRule(rule_data, fp) \
	phgQuadInterfaceDumpRule_(rule_data, fp, NULL, NULL)
#define phgQuadInterfaceGetRuleInfo(rule_data, proj, type) \
	phgQuadInterfaceDumpRule_(rule_data, NULL, proj, type)

/*-------------------------- Tetrahedron functions --------------------------*/

int phgQuadInterface(DOF *ls, DOF *ls_grad, ELEMENT *e,
		DOF_USER_FUNC func, int dim, DOF_PROJ proj,
		int quad_type, int quad_order, FLOAT *res, FLOAT **prule);

#define phgQuadInterfaceGetRule(ls, ls_grad, e, proj, quad_type, quad_order,\
	prule) \
	phgQuadInterface(ls, ls_grad, e, \
		NULL, (proj) == DOF_PROJ_CROSS ? 3 : 1, proj, \
                quad_type, quad_order, NULL, prule)

int phgQuadInterface2(DOF_USER_FUNC ls, int ls_order, DOF_USER_FUNC ls_grad,
		FLOAT tet[Dim + 1][Dim],
		DOF_USER_FUNC func, int dim, DOF_PROJ proj,
		int quad_type, int quad_order, FLOAT *res, FLOAT **prule);

#define phgQuadInterface2GetRule(ls, ls_order, ls_grad, tet, proj, \
	quad_type, quad_order, prule) \
	phgQuadInterface2(ls, ls_order, ls_grad, tet, \
		NULL, 1, proj, quad_type, quad_order, NULL, prule)

INT phgQuadInterfaceMarkElements(DOF *ls);

/*---------------------------- Triangle functions ---------------------------*/

typedef void (*FUNC_2D)(FLOAT x, FLOAT y, FLOAT *value);

void phgQuadInterfaceTriangle(FUNC_2D ls, int ls_order, FUNC_2D ls_grad,
		FLOAT triangle[3][2], FUNC_2D func, int dim, DOF_PROJ proj,
		int quad_type, int quad_order, FLOAT *res, FLOAT **prule);

#define phgQuadInterfaceTriangleGetRule(ls, ls_order, ls_grad, tri, proj, \
	quad_type, quad_order, prule) \
	phgQuadInterfaceTriangle(ls, ls_order, ls_grad, tri, \
		NULL, 1, proj, quad_type, quad_order, NULL, prule)

void phgQuadInterfaceTriangleImportOptions(void);

#ifdef __cplusplus
}
#endif

#define PHG_QUAD_INTERFACE_H
#endif
