/* Parallel Hierarchical Grid -- an adaptive finite element library.
 *
 * Copyright (C) 2005-2010 State Key Laboratory of Scientific and
 * Engineering Computing, Chinese Academy of Sciences. */

/* This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA */

/* Numerical quadrature in a part of a triangle bounded by an implicitly
 * defined interface.
 *
 * Initial code by LIU Ziyang.
 *
 * $Id: quad-interface-triangle.c,v 1.87 2020/07/26 04:04:30 zlb Exp $ */

#include "phg.h"
#include "phg/sturm.h"

#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <setjmp.h>

#undef Dim
#define Dim 2

#define INTER_MAX 20	/* 最大的交点个数 */
#define Opts	_phg_quad_interface_options
static FLOAT D3 = _F(1.0) / _F(3.0);

typedef FLOAT TRIANGLE[Dim + 1][Dim];	/* triangle defined by three vertices */

static struct QuadInfo {
    FUNC_2D	ls;
    FUNC_2D	ls_grad;
    FUNC_2D	u;
    FLOAT	smax, smin, rmax, rmin, s, w_s;
    TRIANGLE	tri;
    QUAD	*quad;
    FLOAT	transform[4], reverse[4];

    BOOLEAN	reverse_flag;
    int		ls_order, dim, quad_type, quad_order, level;
    DOF_PROJ	proj;

    /* variables for recording the quadrature rule */
    FLOAT	*rule;
    int		rule_n, rule_alloc, rule_trunc, rule_size;

    jmp_buf	jmp_env;
} QI;
#if USE_OMP
#pragma omp threadprivate(QI)
#endif /* USE_OMP */

#define LONGJMP if (Opts.subdiv_limit > 0) longjmp(QI.jmp_env, __LINE__)

static FLOAT
max(FLOAT x, FLOAT y)
{
    return x > y ? x : y;
}

static FLOAT
min(FLOAT x, FLOAT y)
{
    return x < y ? x : y;
}

static FLOAT
intersect_seg_seg(const FLOAT *pt1, const FLOAT *pt2,
		  const FLOAT *pt3, const FLOAT *pt4)
/* Returns t in [0,1] such that pt1+t*(pt2-pt1) is the intersection point of
 * the two segments pt1-pt2 and pt3-pt4, or 2.0 if they don't intersect. */
{
    /* 计算线段pt1-pt2与线段pt3-pt4的交点 */
    FLOAT a11, a12, a21, a22, b1, b2;
    FLOAT t, s, d, tol;

    /* We solve the following linear equations:
     *		pt1+t*(pt2-pt1) == pt3+s*(pt4-pt3)
     * for (t,s) in [0,1]^2 */
    a11 = pt2[0] - pt1[0], a12 = pt3[0] - pt4[0],
    a21 = pt2[1] - pt1[1], a22 = pt3[1] - pt4[1],
    b1  = pt3[0] - pt1[0],
    b2  = pt3[1] - pt1[1];

    d = a11 * a22 - a21 * a12;
    if (d == 0.0)
	return 2.0;

    tol = Opts.EPS * d;

    s = a11 * b2 - a21 * b1;
    if ((d > 0. && (s < -tol || s > d + tol)) ||
	(d < 0. && (s > -tol || s < d + tol)))
	return 2.0;

    t = a22 * b1 - a12 * b2;
    if ((d > 0. && (t < -tol || t > d + tol)) ||
	(d < 0. && (t > -tol || t < d + tol)))
	return 2.0;

    t /= d;
    if (t < 0.)
	t = 0.;
    else if (t > 1.)
	t = 1.;

    return t;
}

static int
intersect_seg_triangle(const FLOAT *pt1, const FLOAT *pt2, FLOAT t[3])
/* Returns the number of intersections of the segment pt1-pt2 with the
 * triangle. The positions of the intersections, sorted in ascending order,
 * are returned in t[] (pt1+t*(pt2-pt1)) */
{
    /* 计算线段pt1-pt2与三角形triangled的交点； */
    int n;

    t[0] = intersect_seg_seg(pt1, pt2, QI.tri[0], QI.tri[1]);
    t[1] = intersect_seg_seg(pt1, pt2, QI.tri[0], QI.tri[2]);
    t[2] = intersect_seg_seg(pt1, pt2, QI.tri[1], QI.tri[2]);
    qsort(t, 3, sizeof(t[0]), phgCompFLOAT);
    for (n = 0; n < 3 && t[n] < 1. + Opts.EPS; n++);
    assert(n <= 2);
    return n;
}

static void
unitization(FLOAT *r)
{
    /* 单位化 */
#if 1
    FLOAT d = Sqrt(r[0] * r[0] + r[1] * r[1]);
    assert(d >= Opts.EPS);
    d = 1.0 / d;
#else
    FLOAT d = 1.0 / Sqrt(r[0] * r[0] + r[1] * r[1]);
#endif
    r[0] *= d;
    r[1] *= d;
}

static void
update_QI(void)
{
    FLOAT x1, y1, x2, y2, x0, y0;

    /* 寻找积分方向 */
    QI.ls_grad( (QI.tri[0][0] + QI.tri[1][0] + QI.tri[2][0]) * D3,
		(QI.tri[0][1] + QI.tri[1][1] + QI.tri[2][1]) * D3,
		QI.transform);	/* s方向坐标 */
    if (Fabs(QI.transform[0]) < Opts.EPS && Fabs(QI.transform[1]) < Opts.EPS) {
	/* shouldn't happen if the interface is well resolved */
#if 0
	QI.transform[0] = 1.;
	QI.transform[1] = 0.;
#else
	LONGJMP;
#endif
    }
    unitization(QI.transform);
    QI.transform[2] = -QI.transform[1];
    QI.transform[3] = QI.transform[0];	/* t方向坐标 */

    QI.reverse_flag = FALSE;	/* force recomputing reverse */

    /* 找矩形把三角形区域框起来 */
    x0 = QI.transform[0] * QI.tri[0][0] +
	 QI.transform[1] * QI.tri[0][1];
    y0 = QI.transform[2] * QI.tri[0][0] +
	 QI.transform[3] * QI.tri[0][1];
    x1 = QI.transform[0] * QI.tri[1][0] +
	 QI.transform[1] * QI.tri[1][1];
    y1 = QI.transform[2] * QI.tri[1][0] +
	 QI.transform[3] * QI.tri[1][1];
    x2 = QI.transform[0] * QI.tri[2][0] +
	 QI.transform[1] * QI.tri[2][1];
    y2 = QI.transform[2] * QI.tri[2][0] +
	 QI.transform[3] * QI.tri[2][1];
    QI.rmax = max(x0, max(x1, x2));
    QI.rmin = min(x0, min(x1, x2));
    QI.smax = max(y0, max(y1, y2));
    QI.smin = min(y0, min(y1, y2));

    if (!Opts.show_vectors)
	return;

    fprintf(stderr, "Integration direction:\n  (%0.16lg,%0.16lg)\n",
			(double)QI.transform[0], (double)QI.transform[1]);
    fprintf(stderr, "BoundingBox:\n  %0.16lg %0.16lg %0.16lg %0.16lg\n",
			(double)QI.rmin, (double)QI.smin,
			(double)QI.rmax, (double)QI.smax);
}

static void
user_func(FLOAT x, FLOAT y, const FLOAT *n, const FLOAT w, FLOAT *res)
{
    FLOAT u[QI.dim * Dim];
    int k;

    if (QI.rule_trunc > 0) {
	int pos = RULE_HEADER + QI.rule_n * QI.rule_size;

	if (QI.rule_n >= QI.rule_alloc) {
	    k = QI.rule_alloc + QI.rule_trunc;
	    QI.rule = phgRealloc_(QI.rule,
			    (RULE_HEADER + k * (size_t)QI.rule_size)
					   * sizeof(*QI.rule),
			    (RULE_HEADER + QI.rule_alloc * (size_t)QI.rule_size)
					   * sizeof(*QI.rule));
	    QI.rule_alloc = k;
	}
	/* the coordinates */
	QI.rule[pos++] = x;
	QI.rule[pos++] = y;

	/* the weights */
	if (QI.quad_type != 0 || QI.proj == DOF_PROJ_NONE) {
	    QI.rule[pos++] = w;
	}
	else {
	    QI.rule[pos++] = n[0] * w;
	    QI.rule[pos++] = n[1] * w;
	}
    }

    QI.rule_n++;

    if (res == NULL)
	return;

    if (QI.u == NULL) {
	memset(res, 0, QI.dim * sizeof(*res));
	return;
    }

    if (QI.quad_type != 0 || QI.proj == DOF_PROJ_NONE) {
	QI.u(x, y, res);
	return;
    }

    QI.u(x, y, u);
    switch (QI.proj) {
        case DOF_PROJ_NONE:
	    /* not needed, just to make gcc happy */
	    break;
        case DOF_PROJ_DOT:
	    for (k = 0; k < QI.dim; k++)
	        res[k] = u[Dim * k + 0] * n[0] +
	    		 u[Dim * k + 1] * n[1];
	    break;
        case DOF_PROJ_CROSS:
	    for (k = 0; k < QI.dim; k += 3)
	        res[k] = -u[Dim * k + 0] * n[1] +
	    		  u[Dim * k + 1] * n[0];
	    break;
    }
}

static void
quad_triangle(FLOAT *res)
{
    /* 计算三角形上u(x,y)的积分 */
    QUAD *quad;
    FLOAT *p, *w, s, d, u[QI.dim];	/* Note: proj is not needed here */
    int i, j;
 
    quad = phgQuadGetQuad2D(QI.quad_order);
    p = quad->points;
    w = quad->weights;

    if (res != NULL)
	for (i = 0; i < QI.dim; i++)
	    res[i] = 0.0;

    /* area of the triangle = |det([x0,y0,1; x1,y1,1; x2,y2,1])| / 2! =
	Fabs(x0*y1 + x1*y2 + x2*y0 - y0*x1 - y1*x2 - y2*x0) / 2! */
    s = Fabs(QI.tri[0][0] * QI.tri[1][1] + QI.tri[1][0] * QI.tri[2][1] +
	     QI.tri[2][0] * QI.tri[0][1] - QI.tri[0][1] * QI.tri[1][0] -
	     QI.tri[1][1] * QI.tri[2][0] - QI.tri[2][1] * QI.tri[0][0]) * .5;

    for (j = 0; j < quad->npoints; j++, p += 3, w++) {
	d = s * *w;
	user_func(p[0] * QI.tri[0][0] +
		  p[1] * QI.tri[1][0] +
		  p[2] * QI.tri[2][0],
		  p[0] * QI.tri[0][1] +
		  p[1] * QI.tri[1][1] +
		  p[2] * QI.tri[2][1], NULL, d, u);
	if (res != NULL)
	    for (i = 0; i < QI.dim; i++)
		res[i] += u[i] * d;
    }
}

static int
univariate_poly_roots(int n, FLOAT a, FLOAT b, FLOAT tol,
		      FLOAT *buffer, int **mults)
/* computes the roots in the interval [a,b] of a univariate polynomial of order
 * 'n' and returns the number of roots found.
 *
 * 'buffer'   - Array of length n+1.
 *		On input buffer[i] is the value of the polynomial at the point
 *			a+i*(b-1)/n (i=0,...,n).
 *		On output buffer[k] is the value of the k-th root.
 *
 * 'mults'    -	If not NULL, then it will point to a dynamically allocated
 *		integer array on output (to be freed by the calling function).
 *		with (*mults)[k] equal to the multiplicity of the k-th root.
 */
{
    /* cache the invariant factorized matrix (TODO: cache multiple orders) */
    static BOOLEAN initialized = FALSE;
    static int p = -1;		/* polynomial order */
    static FLOAT *A = NULL;	/* Vandermonde matrix */
    static int *pvt = NULL;	/* pivots of the LU factorization */
#if USE_OMP
#pragma omp threadprivate(p, A, pvt, initialized)
#endif /* USE_OMP */
    int i, j, k, *iwk, n_save;
    FLOAT t, tp, *rwk, *buffer_save;

    if (initialized != TRUE) {
	initialized = TRUE;
	FreeAtExitNoCheck(A);
	FreeAtExitNoCheck(pvt);
    }

    if (p != n) {
	p = n;
	phgFree(A);
	phgFree(pvt);
	A = phgAlloc((p + 1) * (p + 1) * sizeof(*A));
	for (i = 0; i <= p; i++) {
	    tp = 1.0;
	    t = i / (FLOAT)p;
	    for (j = 0; j <= p; j++) {
		A[i * (p + 1) + (p - j)] = tp;
		tp *= t;
	    }
	}
	pvt = phgAlloc((p + 1) * sizeof(*pvt));
	i = phgSolverDenseLU(p + 1, A, pvt);
	assert(i == TRUE);
    }

    phgSolverDenseSV(p + 1, A, pvt, 1, buffer);

    /* Find the largest absolute value of the coefficients */
    t = 0.0;
    for (i = 0; i <= p; i++)
	if (Fabs(buffer[i]) > t)
	    t = Fabs(buffer[i]);
    if (t == 0.0) {
	/* the level-set function is constantly zero along the line,
	 * return the two end-points as solutions with multiplicity 2 */
	*mults = phgAlloc(2 * sizeof(**mults));
	(*mults)[0] = (*mults)[1] = 2;
	buffer[0] = a;
	buffer[1] = b;
	return 2;
    }

    for (i = 0; i <= p; i++) {
	if (Fabs(buffer[i]) > t * Opts.EPS)
	    break;
    }

    assert(i <= p);
    n_save = (n -= i);
    rwk = phgAlloc(((n + 1) * (n + 2) + 4) / 2 * sizeof(*rwk));
    iwk = phgAlloc((n + 1) * sizeof(*iwk));
    buffer_save = phgAlloc((n + 1) * sizeof(*rwk));
    if (i > 0) {
	for (j = 0; i <= p; i++, j++)
	    buffer[j] = buffer[i];
    }
    memcpy(buffer_save, buffer, (n + 1) * sizeof(*buffer));

    n = phgQuadSturm(buffer, n, 0.0, 1.0, tol, Opts.EPS, rwk, iwk);

#if 0
    /* check the solutions */
    for (i = 0; i < n; i++) {
	FLOAT residual, x = buffer[i];
	residual = buffer_save[0];
	for (j = 1; j <= n_save; j++)
	    residual = x * residual + buffer_save[j];
	if (Fabs(residual) > 1e-6) {
	    fprintf(stderr, "p = %d, coeff =", n_save);
	    for (i = 0; i <= n_save; i++)
		fprintf(stderr, " %0.16le", (double)buffer_save[i]);
	    fprintf(stderr, "\nwrong solution: %0.16le, residual: %le\n",
		    (double)x, (double)residual);
	    fprintf(stderr, "abort.\n");
	    exit(1);
	}
    }
#else
    Unused(n_save);
#endif
    phgFree(buffer_save);

    if (n == 0) {
	phgFree(rwk);
	phgFree(iwk);
	if (mults != NULL)
	    *mults = NULL;
	return n;
    }

    /* rescale the roots */
    if (a != 0.0 || b != 1.0) {
	for (i = 0; i < n; i++)
	    rwk[i] = a + buffer[i] * (b - a);
    }
    else {
	for (i = 0; i < n; i++)
	    rwk[i] = buffer[i];
    }

    /* merge very close roots */
    qsort(rwk, n, sizeof(*rwk), phgCompFLOAT);
    for (k = i = 0; i < n; i++, k++) {
	buffer[k] = rwk[i];
	iwk[k] = iwk[i];
	for (j = i + 1; j < n; j++) {
	    tp = Fabs(rwk[j] - rwk[i]);
	    if (tp * tp > Opts.EPS * 10.0)
		break;
	    iwk[k] += iwk[j];
	}
	i = j - 1;
    }
    n = k;

#if 0
    fprintf(stderr, "Number of roots: %d\n", n);
    for (i = 0; i < n; i++)
	fprintf(stderr, "root %d: multiplicity %d, t = %lg\n", i + 1, iwk[i],
		(double)buffer[i]);
#endif

    phgFree(rwk);
    if (mults != NULL)
	*mults = iwk;
    else
	phgFree(iwk);

    return n;
}

static void
shorten_segment(const FLOAT *v0, const FLOAT *v1, FLOAT *t0, FLOAT *t1)
{
    /* 判断v0-v1哪一部分在三角形内部 */
    int n;
    FLOAT t[3];

    n = intersect_seg_triangle(v0, v1, t);
    assert(n == 1 || n == 2);
    switch (n) {
	case 0: *t0 = 1.0; *t1 = 0.0;	break;
	case 1: *t0 = *t1 = t[0];	break;
	case 2: *t0 = t[0]; *t1 = t[1]; break;
    }
}

typedef struct {
    const FLOAT *v0, *v1;
    FLOAT dv[Dim];		/* = v1 - v0 */
} NS_CONTEXT;

static BOOLEAN
newton_func(const FLOAT *t, FLOAT *f, void *pctx)
{
    NS_CONTEXT *nc = pctx;

    QI.ls(nc->v0[0] + (*t) * nc->dv[0], nc->v0[1] + (*t) * nc->dv[1], f);

    return TRUE;
}

static BOOLEAN
newton_jac(const FLOAT *t, FLOAT *jac, void *pctx)
{
    NS_CONTEXT *nc = pctx;
    FLOAT f[Dim];

    QI.ls_grad(nc->v0[0] + (*t) * nc->dv[0],
		    nc->v0[1] + (*t) * nc->dv[1], f);
    *jac = f[0] * nc->dv[0] + f[1] * nc->dv[1];

    return TRUE;
}

static int
intersect_edge(const FLOAT *v0, const FLOAT *v1, BOOLEAN shorten_flag,
	       FLOAT **roots, int **mults)
{
/* 计算线段v0-v1与ls的交点，shorten_flag=TRUE时，只计算在三角形内部的部分 */
    int i, j, order, n;
    FLOAT t0 = 0., t1 = 1., t, *c;
    FLOAT a, b, fa, fb, f;
    BOOLEAN use_newton;
    NS_CONTEXT ctx;

    if (shorten_flag)
	shorten_segment(v0, v1, &t0, &t1);

    if ((t0 -= Opts.EPS) < 0.0)
	t0 = 0.0;
    if ((t1 += Opts.EPS) > 1.0)
	t1 = 1.0;

    if (t0 >= t1) {
	*roots = NULL;
	*mults = NULL;
	return 0;
    }

    order = QI.ls_order;
    use_newton = (Opts.newton == 1 || order < 0);
    order = Opts.newton_porder > 0 ? Opts.newton_porder :
				     (order < 0 ? 2 : order);
    *roots = c = phgAlloc((order + 1) * sizeof(*c));
    for (i = 0; i <= order; i++) {
	t = t0 + i / (FLOAT)order *(t1 - t0);
	QI.ls( v0[0] + t * (v1[0] - v0[0]),
		v0[1] + t * (v1[1] - v0[1]), &c[i] );
    }
    n = univariate_poly_roots(order, t0, t1, Opts.EPS, c, mults);
    ctx.v0 = v0;
    ctx.v1 = v1;
    ctx.dv[0] = v1[0] - v0[0];
    ctx.dv[1] = v1[1] - v0[1];
    j = -1;
    for (i = 0; i < n; i++) {
	int its;
	FLOAT residual;

	if ((*mults)[i] % 2 == 0)
	    continue;

	t = c[i];
	if (use_newton != TRUE) {
	    j++;
	    c[j] = t;
	    (*mults)[j] = 1;
	    continue;
	}

	if (j >= 0 && t <= c[j])
	    continue;

	its = phgNonlinearSolver(1, newton_func, newton_jac, &t,
				 Opts.newton_maxits, Opts.EPS, Opts.EPS,
				 &residual, &ctx);
	if (its >= 0 && its < Opts.newton_maxits) {
	    /* check for duplicate roots */
	    int k;
	    for (k = 0; k <= j; k++)
		if (Fabs(t - c[k]) < Opts.EPS * 10.0)
		    break;
	    if (k <= j)
		continue;	/* skip duplicate root */
	    j++;
	    c[j] = t;
	    (*mults)[j] = 1;
	}
	else {
	    /* try bisection */
	    a = (j < 0 ? t0 : (c[i] + c[j]) * 0.5);
	    b = (i == n - 1 ? t1 : (c[i] + c[i + 1]) * 0.5);
	    assert(a <= c[i] && c[i] <= b);
	    newton_func(&a, &fa, &ctx);
	    newton_func(&b, &fb, &ctx);
	    if (fa * fb > 0.0)
		continue;	/* no root */
	    if (Fabs(fa) == 0.0) {
		t = a;
	    }
	    else if (Fabs(fb) == 0.0) {
		t = b;
	    }
	    else
		while (TRUE) {
		    t = (a + b) * 0.5;
		    if (t == a || t == b || b - a <= Opts.EPS)
			break;
		    newton_func(&t, &f, &ctx);
		    if (Fabs(f) == 0.0)
			break;
		    if ((f > 0.0) == (fa > 0.0)) {
			a = t;
			fa = f;
		    }
		    else {
			b = t;
			fb = f;
		    }
		}
	    j++;
	    c[j] = t;
	    (*mults)[j] = 1;
	}
    }
    n = j + 1;

    return n;
}

static void
reverse_transform(const FLOAT *a, FLOAT *reverse)
{
    /* 矩阵求逆 */
    FLOAT d, det = a[0] * a[3] - a[1] * a[2];
    if (Fabs(det) < Opts.EPS) {
	printf("坐标系转换有误\n");
	exit(0);
    }
    d = 1.0 / det;
    reverse[0] = a[3] * d;
    reverse[1] = -a[1] * d;
    reverse[2] = -a[2] * d;
    reverse[3] = a[0] * d;
}

static void
xyTors(FLOAT x, FLOAT y, FLOAT *r, FLOAT *s, const FLOAT *transform)
{
    /*（x,y）坐标转化为(r,s)坐标 */
    if (r != NULL)
	*r = x * transform[0] + y * transform[1];
    if (s != NULL)
	*s = x * transform[2] + y * transform[3];
}

static void
rsToxy(FLOAT r, FLOAT s, FLOAT *x, FLOAT *y)
{
    /*（r,s）坐标转化为(x,y)坐标 */
    if (!QI.reverse_flag) {
	reverse_transform(QI.transform, QI.reverse);
	QI.reverse_flag = TRUE;
    }
    xyTors(r, s, x, y, QI.reverse);
}

static void
func_f(FLOAT r, FLOAT *res, const FLOAT wgt, void *ctx)
{
    FLOAT x, y;

    Unused(ctx);
    rsToxy(r, QI.s, &x, &y);
    user_func(x, y, NULL, wgt * QI.w_s, res);

    return;
}

static void
func_g(FLOAT s, FLOAT *res, const FLOAT wgt, void *ctx)
{
    FLOAT res0[QI.dim], inter[INTER_MAX + 2];
    FLOAT rpt1[Dim] = { QI.rmin, s }, rpt2[Dim] = { QI.rmax, s };
    FLOAT pt1[Dim], pt2[Dim], r, rmin, rmax, x, y, *roots;
    FLOAT t[3];
    int i, j, n, *mults;

    Unused(ctx);

    if (res != NULL)
	for (i = 0; i < QI.dim; i++)
	    res[i] = 0.0;

    rsToxy(rpt1[0], rpt1[1], &pt1[0], &pt1[1]);
    rsToxy(rpt2[0], rpt2[1], &pt2[0], &pt2[1]);
    i = intersect_seg_triangle(pt1, pt2, t);
    assert(i == 1 || i == 2);
    if (i <= 1)
	return;
    xyTors(pt1[0] + t[0] * (pt2[0] - pt1[0]),
	   pt1[1] + t[0] * (pt2[1] - pt1[1]), &rmin, NULL, QI.transform);
    xyTors(pt1[0] + t[1] * (pt2[0] - pt1[0]),
	   pt1[1] + t[1] * (pt2[1] - pt1[1]), &rmax, NULL, QI.transform);
    if (rmin > rmax) {
	r = rmin;
	rmin = rmax;
	rmax = r;
    }
    if (rmax - rmin <= Opts.EPS + Opts.EPS)
	return;

    n = 0;
    inter[n++] = rmin;
    i = intersect_edge(pt1, pt2, TRUE, &roots, &mults);
    if (i >= 2 || (i == 1 && mults[0] > 1)) {
	phgFree(roots);
	phgFree(mults);
	LONGJMP;
    }
    phgFree(mults);
    if (i > 0) {
	xyTors( pt1[0] + roots[0] * (pt2[0] - pt1[0]),
		pt1[1] + roots[0] * (pt2[1] - pt1[1]), &r, NULL, QI.transform);
	if (r > rmin + Opts.EPS && r < rmax - Opts.EPS)
	    inter[n++] = r;
    }
    inter[n++] = rmax;
    phgFree(roots);

    QI.s = s;
    QI.w_s = wgt;
    for (i = 0; i < n - 1; i++) {
	FLOAT ls;
	/* the interval is entirely either inside or outside the subdomain,
	 * we simply check the middle point of the interval to decide whether
	 * to skip it. */
	rsToxy(0.5 * (inter[i] + inter[i + 1]), s, &x, &y);
	QI.ls(x, y, &ls);
	if ((QI.quad_type < 0 ? ls : -ls) > 0.)
	    continue;
	phgQuad1D(func_f, QI.dim, inter[i], inter[i + 1], QI.quad,
		  res == NULL ? NULL : res0, NULL);
	if (res != NULL)
	    for (j = 0; j < QI.dim; j++)
		res[j] += res0[j];
    }

    return;
}

static void
func_g_curve(FLOAT s, FLOAT *res, const FLOAT wgt, void *ctx)
{
    int *mults, i, n;
    FLOAT v0[Dim], v1[Dim], *roots, pt[Dim];

    Unused(ctx);

    rsToxy(QI.rmin, s, &v0[0], &v0[1]);
    rsToxy(QI.rmax, s, &v1[0], &v1[1]);
    n = intersect_edge(v0, v1, TRUE, &roots, &mults);
    if (n <= 0) {
	if (res != NULL)
	    for (i = 0; i < QI.dim; i++)
		res[i] = 0.;
	phgFree(roots);
	phgFree(mults);
	return;
    }
    else if (n >= 2 || mults[0] > 1) {
	phgFree(roots);
	phgFree(mults);
	LONGJMP;
    }
    else {
	FLOAT grad[Dim], gradrs[Dim], d;
	pt[0] = v0[0] + (*roots) * (v1[0] - v0[0]);
	pt[1] = v0[1] + (*roots) * (v1[1] - v0[1]);
	QI.ls_grad(pt[0], pt[1], grad);
	if (!QI.reverse_flag) {
	    reverse_transform(QI.transform, QI.reverse);
	    QI.reverse_flag = TRUE;
	}
	gradrs[0] = grad[0] * QI.reverse[0] + grad[1] * QI.reverse[2];
	gradrs[1] = grad[0] * QI.reverse[1] + grad[1] * QI.reverse[3];
	d = Sqrt(1. + (gradrs[1] * gradrs[1]) / (gradrs[0] * gradrs[0]));
	if (QI.proj != DOF_PROJ_NONE)
	    unitization(grad);
	user_func(pt[0], pt[1], grad, d * wgt, res);
	for (i = 0; i < QI.dim; i++)
	    res[i] *= d;
    }
    phgFree(mults);
    phgFree(roots);
    return;
}

/*******************************************************************************************************************************************/

static void quad_interface_triangle(FLOAT *res);

static void
subdivide_triangle(FLOAT *res)
{
    FLOAT pt1[Dim] = {(QI.tri[0][0] + QI.tri[1][0]) * 0.5,
		      (QI.tri[0][1] + QI.tri[1][1]) * 0.5};
    FLOAT pt2[Dim] = {(QI.tri[0][0] + QI.tri[2][0]) * 0.5,
		      (QI.tri[0][1] + QI.tri[2][1]) * 0.5};
    FLOAT pt3[Dim] = {(QI.tri[2][0] + QI.tri[1][0]) * 0.5,
		      (QI.tri[2][1] + QI.tri[1][1]) * 0.5};
    FLOAT res0[QI.dim];
    TRIANGLE triangles[4];
    int i, j;

    QI.level++;

    triangles[0][0][0] = QI.tri[0][0];
    triangles[0][0][1] = QI.tri[0][1];
    triangles[0][1][0] = pt1[0];
    triangles[0][1][1] = pt1[1];
    triangles[0][2][0] = pt2[0];
    triangles[0][2][1] = pt2[1];

    triangles[1][0][0] = QI.tri[1][0];
    triangles[1][0][1] = QI.tri[1][1];
    triangles[1][1][0] = pt1[0];
    triangles[1][1][1] = pt1[1];
    triangles[1][2][0] = pt3[0];
    triangles[1][2][1] = pt3[1];

    triangles[2][0][0] = QI.tri[2][0];
    triangles[2][0][1] = QI.tri[2][1];
    triangles[2][1][0] = pt3[0];
    triangles[2][1][1] = pt3[1];
    triangles[2][2][0] = pt2[0];
    triangles[2][2][1] = pt2[1];

    triangles[3][0][0] = pt3[0];
    triangles[3][0][1] = pt3[1];
    triangles[3][1][0] = pt1[0];
    triangles[3][1][1] = pt1[1];
    triangles[3][2][0] = pt2[0];
    triangles[3][2][1] = pt2[1];

    memcpy(QI.tri, triangles[0], sizeof(QI.tri));
    quad_interface_triangle(res);
    for (j = 1; j < 4; j++) {
	memcpy(QI.tri, triangles[j], sizeof(QI.tri));
	quad_interface_triangle(res == NULL ? NULL : res0);
	if (res != NULL)
	    for (i = 0; i < QI.dim; i++)
		res[i] += res0[i];
    }

    QI.level--;

    return;
}

static void
quad_interface_triangle(FLOAT *res)
{
/* 计算三角形triangle和曲线ls相交区域上的积分 */
    FLOAT s, *roots, thres;
    int i, j, k, n, rule_n, *mults;

    /* variables for recording intersections */
    int sinter_n;
    /* use a static buffer to save memory with regard to recursion */
    static FLOAT sinter[INTER_MAX];
#if USE_OMP
#pragma omp threadprivate(sinter)
#endif /* USE_OMP */

    if (QI.level < Opts.subdiv_level0) {
	/* This is the 1st of the 2 entries for subdividing the element */
	subdivide_triangle(res);
	return;
    }

    rule_n = QI.rule_n;	/* save position */
    if ((i = setjmp(QI.jmp_env)) != 0) {
	/* This is the 2nd of the 2 entries for subdividing the element */
	QI.rule_n = rule_n;	/* restore position of rule_data */
	if (Opts.show_recursions == TRUE)
	    fprintf(stderr, "Recursion from line %d, level %d\n", i, QI.level);
	subdivide_triangle(res);
	return;
    }

    if (QI.level > Opts.subdiv_limit)
	phgError(1, "%s: recursion limit exceeded, abort.\n", __func__);

    /* update 'QI' to match the current triangle */
    update_QI();

    /* Construct integration intervals for s */

    /* Note: cos(a) > Opts.threshold <==> sin(a)^2 < 1 - Opts.threshold^2 */
    thres = 1.0 - Opts.threshold * Opts.threshold;

    /* First, add projections of the intersections on the 3 edges */
    n = 0;
    for (i = 0; i < 3; i++) {	/* loop on the 3 egdes */
	int v0 = i, v1 = (i == 2 ? 0 : i + 1);
	j = intersect_edge(QI.tri[v0], QI.tri[v1], FALSE, &roots, &mults);
	assert(n + j <= INTER_MAX);
	for (k = 0; k < j; k++) {
	    FLOAT x, y, pt[Dim];
	    if (mults[k] > 1)
		LONGJMP;
	    x = QI.tri[v0][0] + roots[k] * (QI.tri[v1][0] - QI.tri[v0][0]);
	    y = QI.tri[v0][1] + roots[k] * (QI.tri[v1][1] - QI.tri[v0][1]);
	    QI.ls_grad(x, y, pt);
	    /* check angle between tangent vector and r direction */
	    s = QI.transform[0] * pt[0] + QI.transform[1] * pt[1];
	    /* Note: s*s/|pt|^2 = sin(theta)^2 */
	    if (s * s < thres * (pt[0] * pt[0] + pt[1] * pt[1]))
		LONGJMP; /* integr diretion too parallel to the interface */
	    xyTors(x, y, NULL, &s, QI.transform);
	    sinter[n++] = s;
	}
	phgFree(roots);
	phgFree(mults);
    }

#if 0
    /* Can't do this if the interface is a closed curve inside the triangle */
    if (n == 0) {
	FLOAT ls1, ls2, ls0;
	/* The interface does not intersect with the 3 edges */
	if (QI.quad_type == 0) {
    zero_res:
	    if (res != NULL)
		for (k = 0; k < QI.dim; k++)
		    res[k] = 0.;
	    return;
	}
	/* check if the triangle is within the integration domain */
	QI.ls(QI.tri[0][0], QI.tri[0][1], &ls0);
	QI.ls(QI.tri[1][0], QI.tri[1][1], &ls1);
	QI.ls(QI.tri[2][0], QI.tri[2][1], &ls2);
	if (QI.quad_type > 0) {
	    ls0 = -ls0;
	    ls1 = -ls1;
	    ls2 = -ls2;
	}
	if (ls0 >= Opts.EPS || ls1 >= Opts.EPS || ls2 >= Opts.EPS)
	    goto zero_res;
	/* the whole triangle is in the integration domain */
	quad_triangle(res);
	return;
    }
#endif

    /* Next, add projections of the 3 vertices */
    assert(n + 3 <= INTER_MAX);
    for (i = 0; i < 3; i++) {
	xyTors(QI.tri[i][0], QI.tri[i][1], NULL, &s, QI.transform);
	sinter[n++] = s;
    }

    sinter_n = n;

    /* sort sinter[] */
    qsort(sinter, n, sizeof(sinter[0]), phgCompFLOAT);

    /* remove duplicate entries in sinter[] */
    for (i = 0, j = 1; j < sinter_n; ) {
	while (j < sinter_n && Fabs(sinter[j] - sinter[i]) < Opts.EPS)
	    j++;
	if (j >= sinter_n) 
	    break;
	sinter[++i] = sinter[j];
    }
    sinter_n = ++i;

    if (QI.quad_type != 0) {
	FLOAT res0[QI.dim];
	if (res != NULL)
	    for (k = 0; k < QI.dim; k++)
		res[k] = 0.;
	for (i = 0; i < sinter_n - 1; i = j) {
	    for (j = i + 1; j < sinter_n && sinter[j] - sinter[i] <= Opts.EPS;
									j++);
	    if (j >= sinter_n)
		break;
	    phgQuad1D(func_g, QI.dim, sinter[i], sinter[j],
			  QI.quad, res == NULL ? NULL : res0, NULL);
	    if (res != NULL && Opts.show_intervals_s && !Opts.dbg_s)
		printf("s_interval (%0.16g,%0.16g), res0=%0.16lg\n",
			(double)sinter[i], (double)sinter[j], (double)res0[0]);
	    if (res != NULL && Opts.dbg_s) {
		static FLOAT error0 = 0.0;
#if USE_OMP
# pragma omp threadprivate(error0)
#endif  /* USE_OMP */
		FLOAT error, res1[QI.dim];
		phgQuad1D(func_g, QI.dim, sinter[i], sinter[j],
			  phgQuadGetQuad1D(QI.quad_order + 2), res1, NULL);
		error = Fabs((res0[0] - res1[0]) /
				(res1[0] == 0. ? 1. : res1[0]));
		if (error > error0 || Opts.show_intervals_s) {
		    printf("s_interval (%0.16g,%0.16g), res0=%0.16g, err=%e\n",
				(double)sinter[i], (double)sinter[j],
				(double)res0[0], (double)error);
		    if (error > error0)
			error0 = error;
		}
	    }
	    if (res != NULL)
		for (k = 0; k < QI.dim; k++)
		    res[k] += res0[k];
	}
#if 0
        fprintf(stderr, "tri: (%lg,%lg),(%lg,%lg),(%lg,%lg), res = %lg\n",
			(double)QI.tri[0][0], (double)QI.tri[0][1],
			(double)QI.tri[1][0], (double)QI.tri[1][1],
			(double)QI.tri[2][0], (double)QI.tri[2][1],
			(double)*res);
#endif
    }
    else {
	FLOAT res0[QI.dim];
	if (res != NULL)
	    for (k = 0; k < QI.dim; k++)
		res[k] = 0.;
	for (i = 0; i < sinter_n - 1; i = j) {
	    for (j = i + 1; j < sinter_n && sinter[j] - sinter[i] <= Opts.EPS;
									j++);
	    if (j >= sinter_n)
		break;
	    phgQuad1D(func_g_curve, QI.dim, sinter[i], sinter[j],
			  QI.quad, res == NULL ? NULL : res0, NULL);
	    if (res != NULL)
		for (k = 0; k < QI.dim; k++)
		    res[k] += res0[k];
	}
    }

    if (res != NULL && Opts.dbg_elem && res[0] != 0.0) {
	static double error = 0.0;
	FLOAT err, res0[QI.dim];
	QUAD *quad_save = QI.quad;
	QI.quad = phgQuadGetQuad1D(QI.quad_order + 2);
	Opts.dbg_elem = FALSE;
	quad_interface_triangle(res0);
	Opts.dbg_elem = TRUE;
	err = Fabs(1.0 - res0[0] / res[0]);
	if (error < err) {
	    error = err;
	    fprintf(stderr, "Elem={{%0.16lg,%0.16lg},{%0.16lg,%0.16lg},"
		    "{%0.16lg,%0.16lg}}, res=%lg, res0=%lg, error=%le\n",
		    (double)QI.tri[0][0], (double)QI.tri[0][1],
		    (double)QI.tri[1][0], (double)QI.tri[1][1],
		    (double)QI.tri[2][0], (double)QI.tri[2][1],
		    (double)res[0], (double)res0[0], error);
	}
	QI.quad = quad_save;
    }
}

void
phgQuadInterfaceTriangle(FUNC_2D ls, int ls_order, FUNC_2D ls_grad,
			 TRIANGLE triangle, FUNC_2D u, int dim, DOF_PROJ proj,
			 int quad_type, int quad_order, FLOAT *res,
			 FLOAT **prule_data)
/* See comments in the function phgQuadInterface (in quad-interface.c),
 * or doc/quad-XFEM.doc for documentations about the arguments. */
{
    int i, j;

    if (quad_type != 0)
	proj = DOF_PROJ_NONE;

    QI.ls = ls;
    QI.ls_order = ls_order;
    QI.ls_grad = ls_grad;
    QI.u = u;
    QI.dim = dim;	/* dimension of 'res' */
    QI.proj = proj;
    for (i = 0; i < Dim + 1; i++) {
	for (j = 0; j < Dim; j++)
	    QI.tri[i][j] = triangle[i][j];
    }
    QI.quad = phgQuadGetQuad1D(quad_order);
    QI.quad_order = quad_order;
    QI.quad_type = quad_type;
    QI.level = 0;

#if 0
#warning test-only code!
    if (quad_type > 0) {
	FLOAT res0[dim];
	/* Test: compute integral as the complement of quad_type < 0 */
	memcpy(QI.tri, triangle, sizeof(QI.tri));
	quad_triangle(res);
	phgQuadInterfaceTriangle(ls, ls_order, ls_grad, triangle, u, dim,
				 proj, -1, quad_order, res0, NULL);
	for (i = 0; i < dim; i++)
	    res[i] -= res0[i];
	return;
    }
#endif

    QI.rule_n = QI.rule_alloc = 0;
    QI.rule = NULL;
    if (prule_data == NULL && Opts.dbg_rule_data == FALSE) {
	QI.rule_trunc = QI.rule_size = 0;
    }
    else {
	int np = (quad_order-1)/2+1;	/* # points of the 1D Gaussian rule */
	QI.rule_trunc = 10 * np;
	QI.rule_size = Dim + 1;		/* x,y, w_{1x1} */
	if (quad_type != 0)
	    QI.rule_trunc *= 2 * np;
	else if (proj == DOF_PROJ_DOT || proj == DOF_PROJ_CROSS)
	    QI.rule_size += Dim - 1;	/* weight * unit normal */
	QI.rule = phgAlloc((RULE_HEADER + 
			(QI.rule_alloc = QI.rule_trunc) * QI.rule_size)
					 * sizeof(*QI.rule));
    }

    if (ls == NULL) {
	if (quad_type != 0)
	    quad_triangle(res);
	else if (res != NULL)
	    memset(res, 0, dim * sizeof(*res));
    }
    else {
	quad_interface_triangle(res);
    }

    if (QI.rule_n == 0) {
	phgFree(QI.rule);
	QI.rule = NULL;
    }
    else if (QI.rule != NULL) {
	SetRuleHeader(QI.rule, QI.rule_n, proj, 0);
    }

    if (Opts.dbg_rule_data != FALSE) {
#if 0
#warning remove me!
	j = phgQuadInterfaceGetRuleInfo(QI.rule, &proj, &i);
	assert(j == 0 || (proj == QI.proj && i == 0 && j == QI.rule_n));
	printf("resulting quadrature rule: %d points\n", j);
#endif
	if (u == NULL) {
	    phgPrintf("Note: -qi_dbg_rule_data ignored since u == NULL!\n");
	}
	else if (QI.rule_n > 0) {
	    FLOAT res0[dim];
	    assert(QI.rule != NULL);
	    phgQuadInterfaceApplyRule(QI.u, QI.dim, QI.rule, res0);
	    for (i = 0; i < dim; i++) {
		FLOAT err = (res[i] - res0[i]) / (res[i] == 0.0 ? 1.0 : res[i]);
#if 0
#warning remove me!
		printf("%d: res = %0.16lg, err = %le\n",
					i, (double)res[i], (double)err);
#endif
		if (Fabs(err) > 1000.0 * Opts.EPS)
		    phgError(1, "rule checking failed: %lg<=>%lg, err=%le\n",
				(double)res[i], (double)res0[i], (double)err);
	    }
	}
    }

    if (prule_data != NULL)
	*prule_data = QI.rule;
    else
	phgFree(QI.rule);
}
